@@ -75,6 +75,9 @@
 #if HAVE_SYS_VFS_H
 #include <sys/vfs.h>
 #elif HAVE_SYS_MOUNT_H
+#if HAVE_SYS_PARAM_H
+#include <sys/param.h>
+#endif
 #include <sys/mount.h>
 #endif
 
@@ -90,14 +93,19 @@
 #define DIR_SEPARATORS "/"
 #endif
 
+typedef struct {
+	GnomeVFSMethodMonitorCancelFunc cancel_func;  /* Must be first */
+} AnyFileMonitorHandle;
+
 #ifdef HAVE_FAM
 static FAMConnection *fam_connection = NULL;
 static gint fam_watch_id = 0;
 G_LOCK_DEFINE_STATIC (fam_connection);
 
 typedef struct {
-	FAMRequest request;
+	GnomeVFSMethodMonitorCancelFunc cancel_func;  /* Must be first */
 	GnomeVFSURI *uri;
+	FAMRequest request;
 	gboolean     cancelled;
 } FileMonitorHandle;
 #endif
@@ -1166,11 +1174,7 @@ do_get_file_info (GnomeVFSMethod *method,
 	}
 
 	if (options & GNOME_VFS_FILE_INFO_GET_SELINUX_CONTEXT) {
-		result = get_selinux_context (file_info, full_name, options);
-		if (result != GNOME_VFS_OK) { 
-			g_free (full_name);
-			return result;
-		}
+		get_selinux_context (file_info, full_name, options);
 	} 
 
 	if (options & GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS) {
@@ -1242,16 +1246,17 @@ do_get_file_info_from_handle (GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
-GHashTable *fstype_hash = NULL;
-G_LOCK_DEFINE_STATIC (fstype_hash);
 extern char *filesystem_type (char *path, char *relpath, struct stat *statp);
 
+G_LOCK_DEFINE_STATIC (fstype);
+
 static gboolean
 do_is_local (GnomeVFSMethod *method,
 	     const GnomeVFSURI *uri)
 {
+	struct stat statbuf;
+	gboolean is_local;
 	gchar *path;
-	gpointer local = NULL;
 
 	g_return_val_if_fail (uri != NULL, FALSE);
 
@@ -1259,30 +1264,23 @@ do_is_local (GnomeVFSMethod *method,
 	if (path == NULL)
 		return TRUE; /* GNOME_VFS_ERROR_INVALID_URI */
 
-	G_LOCK (fstype_hash);
-	if (fstype_hash == NULL)
-		fstype_hash = g_hash_table_new_full (
-			g_str_hash, g_str_equal, g_free, NULL);
-	else
-		local = g_hash_table_lookup (fstype_hash, path);
-
-	if (local == NULL) {
-		struct stat statbuf;
-		if (g_stat (path, &statbuf) == 0) {
-			char *type = filesystem_type (path, path, &statbuf);
-			gboolean is_local = ((strcmp (type, "nfs") != 0) && 
-					     (strcmp (type, "afs") != 0) &&
-					     (strcmp (type, "autofs") != 0) &&
-					     (strcmp (type, "unknown") != 0) &&
-					     (strcmp (type, "ncpfs") != 0));
-			local = GINT_TO_POINTER (is_local ? 1 : -1);
-			g_hash_table_insert (fstype_hash, path, local);
-		}
-	} else
-		g_free (path);
-
-	G_UNLOCK (fstype_hash);
-	return GPOINTER_TO_INT (local) > 0;
+	if (g_stat (path, &statbuf) == 0) {
+		char *type;
+		
+		G_LOCK (fstype);
+		type = filesystem_type (path, path, &statbuf);
+		is_local = ((strcmp (type, "nfs") != 0) && 
+			    (strcmp (type, "afs") != 0) &&
+			    (strcmp (type, "autofs") != 0) &&
+			    (strcmp (type, "unknown") != 0) &&
+			    (strcmp (type, "ncpfs") != 0));
+		G_UNLOCK (fstype);
+	} else {
+		/* Assume non-existent files are local */
+		is_local = TRUE;
+	}
+	g_free (path);
+	return is_local;
 }
 
 
@@ -2136,6 +2134,41 @@ rename_helper (const gchar *old_full_name,
 	return GNOME_VFS_OK;
 }
 
+static GnomeVFSResult
+set_symlink_name_helper (const gchar *full_name,
+			 const GnomeVFSFileInfo *info)
+{
+#ifndef G_OS_WIN32
+	struct stat statbuf;
+
+	if (info->symlink_name == NULL) {
+		return GNOME_VFS_ERROR_BAD_PARAMETERS;
+	}
+
+	if (g_lstat (full_name, &statbuf) != 0) {
+		return gnome_vfs_result_from_errno ();
+	}
+
+	if (!S_ISLNK (statbuf.st_mode)) {
+		return GNOME_VFS_ERROR_NOT_A_SYMBOLIC_LINK;
+	}
+
+	if (g_unlink (full_name) != 0) {
+		return gnome_vfs_result_from_errno ();
+	}
+
+	if (symlink (info->symlink_name, full_name) != 0) {
+		/* TODO should we try to restore the old symlink? */
+		return gnome_vfs_result_from_errno ();
+	}
+
+	return GNOME_VFS_OK;
+#else
+	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+#endif
+
+}
+
 static GnomeVFSResult
 do_move (GnomeVFSMethod *method,
 	 GnomeVFSURI *old_uri,
@@ -2397,7 +2430,16 @@ do_set_file_info (GnomeVFSMethod *method,
 			return result;	
 		}
 	}
-	
+
+	if (mask & GNOME_VFS_SET_FILE_INFO_SYMLINK_NAME) {
+		GnomeVFSResult result;
+
+		result = set_symlink_name_helper (full_name, info);
+		if (result != GNOME_VFS_OK) {
+			g_free (full_name);
+			return result;	
+		}
+	}
 
 	g_free (full_name);
 
@@ -2532,102 +2574,9 @@ monitor_setup (void)
 #endif
 
 #ifdef HAVE_FAM
-static GnomeVFSResult fam_monitor_add (GnomeVFSMethod *method,
-				       GnomeVFSMethodHandle **method_handle_return,
-				       GnomeVFSURI *uri,
-				       GnomeVFSMonitorType monitor_type)
-{
-	FileMonitorHandle *handle;
-	char *filename;
-
-	if (!monitor_setup ()) {
-		return GNOME_VFS_ERROR_NOT_SUPPORTED;
-	}
-
-	filename = get_path_from_uri (uri);
-	if (filename == NULL) {
-		return GNOME_VFS_ERROR_INVALID_URI;
-	}
-	
-	handle = g_new0 (FileMonitorHandle, 1);
-	handle->uri = uri;
-	handle->cancelled = FALSE;
-	gnome_vfs_uri_ref (uri);
-
-	G_LOCK (fam_connection);
-	/* We need to queue up incoming messages to avoid blocking on write
-	   if there are many monitors being added */
-	fam_do_iter_unlocked ();
-
-	if (fam_connection == NULL) {
-		G_UNLOCK (fam_connection);
-		return GNOME_VFS_ERROR_NOT_SUPPORTED;
-	}
-	
-	if (monitor_type == GNOME_VFS_MONITOR_FILE) {
-		FAMMonitorFile (fam_connection, filename, 
-			&handle->request, handle);
-	} else {
-		FAMMonitorDirectory (fam_connection, filename, 
-			&handle->request, handle);
-	}
-
-	G_UNLOCK (fam_connection);
-	
-	*method_handle_return = (GnomeVFSMethodHandle *)handle;
-
-	g_free (filename);
-
-	return GNOME_VFS_OK;
-
-}
-#endif
-
-#ifdef USE_INOTIFY
-static GnomeVFSResult inotify_monitor_add (GnomeVFSMethod *method,
-					   GnomeVFSMethodHandle **method_handle_return,
-					   GnomeVFSURI *uri,
-					   GnomeVFSMonitorType monitor_type)
-{
-	ih_sub_t *sub;
-
-	sub = ih_sub_new (uri, monitor_type);
-	if (sub == NULL) {
-		return GNOME_VFS_ERROR_INVALID_URI;
-	}
-
-	if (ih_sub_add (sub) == FALSE) {
-		ih_sub_free (sub);
-		*method_handle_return = NULL;
-		return GNOME_VFS_ERROR_INVALID_URI;
-	}
-
-	*method_handle_return = (GnomeVFSMethodHandle *)sub;
-	return GNOME_VFS_OK;
-
-}
-#endif
-
 static GnomeVFSResult
-do_monitor_add (GnomeVFSMethod *method,
-		GnomeVFSMethodHandle **method_handle_return,
-		GnomeVFSURI *uri,
-		GnomeVFSMonitorType monitor_type)
-{
-#ifdef USE_INOTIFY
-	if (ih_startup ()) {
-		return inotify_monitor_add (method, method_handle_return, uri, monitor_type);
-	}
-#endif
-#ifdef HAVE_FAM
-	return fam_monitor_add (method, method_handle_return, uri, monitor_type);
-#endif
-	return GNOME_VFS_ERROR_NOT_SUPPORTED;
-}
-
-#ifdef HAVE_FAM
-static GnomeVFSResult fam_monitor_cancel (GnomeVFSMethod *method,
-					  GnomeVFSMethodHandle *method_handle)
+fam_monitor_cancel (GnomeVFSMethod *method,
+		    GnomeVFSMethodHandle *method_handle)
 {
 	FileMonitorHandle *handle = (FileMonitorHandle *)method_handle;
 
@@ -2655,12 +2604,69 @@ static GnomeVFSResult fam_monitor_cancel (GnomeVFSMethod *method,
 
 	return GNOME_VFS_OK;
 }
+
+
+static GnomeVFSResult
+fam_monitor_add (GnomeVFSMethod *method,
+		 GnomeVFSMethodHandle **method_handle_return,
+		 GnomeVFSURI *uri,
+		 GnomeVFSMonitorType monitor_type)
+{
+	FileMonitorHandle *handle;
+	char *filename;
+
+	if (!monitor_setup ()) {
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
+
+	filename = get_path_from_uri (uri);
+	if (filename == NULL) {
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
+	
+	handle = g_new0 (FileMonitorHandle, 1);
+	handle->cancel_func = fam_monitor_cancel;
+	handle->uri = uri;
+	handle->cancelled = FALSE;
+	gnome_vfs_uri_ref (uri);
+
+	G_LOCK (fam_connection);
+	/* We need to queue up incoming messages to avoid blocking on write
+	   if there are many monitors being added */
+	fam_do_iter_unlocked ();
+
+	if (fam_connection == NULL) {
+		G_UNLOCK (fam_connection);
+		g_free (handle);
+		gnome_vfs_uri_unref (uri);
+		g_free (filename);
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
+	
+	if (monitor_type == GNOME_VFS_MONITOR_FILE) {
+		FAMMonitorFile (fam_connection, filename, 
+			&handle->request, handle);
+	} else {
+		FAMMonitorDirectory (fam_connection, filename, 
+			&handle->request, handle);
+	}
+
+	G_UNLOCK (fam_connection);
+	
+	*method_handle_return = (GnomeVFSMethodHandle *)handle;
+
+	g_free (filename);
+
+	return GNOME_VFS_OK;
+
+}
 #endif
 
-
 #ifdef USE_INOTIFY
-static GnomeVFSResult inotify_monitor_cancel (GnomeVFSMethod *method,
-					      GnomeVFSMethodHandle *method_handle)
+
+static GnomeVFSResult
+inotify_monitor_cancel (GnomeVFSMethod *method,
+			GnomeVFSMethodHandle *method_handle)
 {
 	ih_sub_t *sub = (ih_sub_t *)method_handle;
 
@@ -2671,21 +2677,61 @@ static GnomeVFSResult inotify_monitor_cancel (GnomeVFSMethod *method,
 	ih_sub_free (sub);
 	return GNOME_VFS_OK;
 
+}
+
+static GnomeVFSResult
+inotify_monitor_add (GnomeVFSMethod *method,
+		     GnomeVFSMethodHandle **method_handle_return,
+		     GnomeVFSURI *uri,
+		     GnomeVFSMonitorType monitor_type)
+{
+	ih_sub_t *sub;
+
+	sub = ih_sub_new (uri, monitor_type);
+	if (sub == NULL) {
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
+	sub->cancel_func = inotify_monitor_cancel;
+	if (ih_sub_add (sub) == FALSE) {
+		ih_sub_free (sub);
+		*method_handle_return = NULL;
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
+
+	*method_handle_return = (GnomeVFSMethodHandle *)sub;
+	return GNOME_VFS_OK;
+
 }
 #endif
 
+static GnomeVFSResult
+do_monitor_add (GnomeVFSMethod *method,
+		GnomeVFSMethodHandle **method_handle_return,
+		GnomeVFSURI *uri,
+		GnomeVFSMonitorType monitor_type)
+{
+#ifdef USE_INOTIFY
+	/* For remote files, always prefer FAM */
+	if (do_is_local (method, uri) && ih_startup ()) {
+		return inotify_monitor_add (method, method_handle_return, uri, monitor_type);
+	}
+#endif
+#ifdef HAVE_FAM
+	return fam_monitor_add (method, method_handle_return, uri, monitor_type);
+#endif
+	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+}
+
 static GnomeVFSResult
 do_monitor_cancel (GnomeVFSMethod *method,
 		   GnomeVFSMethodHandle *method_handle)
 {
-#ifdef USE_INOTIFY
-	if (ih_startup ()) {
-		return inotify_monitor_cancel (method, method_handle);
+	AnyFileMonitorHandle *handle;
+
+	handle = (AnyFileMonitorHandle *)method_handle;
+	if (handle != NULL) {
+		return (handle->cancel_func) (method, method_handle);
 	}
-#endif
-#ifdef HAVE_FAM
-	return fam_monitor_cancel (method, method_handle);
-#endif
 	return GNOME_VFS_ERROR_NOT_SUPPORTED;
 }
 
